import os
import glob
import shutil
from .path import join_path,get_relative_path
def is_empty_dir(path):
    return os.path.exists(path) and os.path.isdir(path) and len(os.listdir(path))==0
def clean_dir(path):
    assert os.path.isdir(path)
    for item in os.listdir(path):
        child_path=os.path.join(path,item)
        if os.path.isdir(child_path):
            shutil.rmtree(child_path)
        else:
            os.remove(child_path)
def check_and_make_empty_dir(dst,overwrite):
    if os.path.exists(dst):
        if os.path.isfile(dst):
            if overwrite:
                os.remove(dst)
                os.makedirs(dst)
            raise NotADirectoryError('Not a directory:%s'%(dst))
        elif not is_empty_dir(dst):
            if overwrite:
                clean_dir(dst)
            else:
                raise RuntimeError('Directory nor empty : %s'%(dst))
    else:
        os.makedirs(dst)
def generate_hash(s,times=1):
    assert times>=1
    import hashlib
    m = hashlib.md5()
    def gen():
        m.update(s.encode('utf-8'))
        return m.hexdigest()[:10]
    for i in range(times):
        data=gen()
    return data



def copy_repo_files_to(repo_path,relative_path,dst_path,repo_name):
    '''

    :param src:
    :param dst:
    :return:
    TODO:
        should overwrite if file exists?
    '''
    src_path=join_path(repo_path,relative_path)
    if os.path.isfile(src_path):
        shutil.copy(src_path,dst_path)
    elif '*' in src_path:
        # dst_path must be a dir
        items=glob.glob(src_path,recursive=True)
        if not os.path.exists(dst_path):
            os.makedirs(dst_path)
        assert os.path.isdir(dst_path)
        for item_path in items:
            item_rel_path=get_relative_path(repo_path,item_path)
            item_dst_path=join_path(dst_path,item_rel_path)
            if os.path.isdir(item_path):
                shutil.copytree(item_path,item_dst_path)
            else:
                assert os.path.isfile(item_path)
                shutil.copy(item_path,item_dst_path)
    else:
        assert os.path.isdir(src_path)
        if os.path.exists(dst_path):
            assert os.path.isdir(dst_path)
            dst_name=repo_name if relative_path=='/' else os.path.basename(src_path)
            dst_path=join_path(dst_path,dst_name)
            os.makedirs(dst_path)
        else:
            os.makedirs(dst_path)
        if relative_path=='/':
            # copy the whole repo
            copy_repo_files_to_dir(src_path,dst_path,exclude_git=True)
        else:
            # copy some dir under the repo
            copy_repo_files_to_dir(src_path,dst_path,exclude_git=False)


def copy_repo_files_to_dir(src,dst,exclude_git=True):
    if not os.path.exists(dst):
        os.makedirs(dst)
    assert os.path.isdir(dst)
    for child in os.listdir(src):
        if child=='.git' and exclude_git:
            continue
        child_path=os.path.join(src,child)
        dst_child_path=os.path.join(dst,child)
        if os.path.isdir(child_path):
            shutil.copytree(child_path,dst_child_path)
        else:
            shutil.copy(child_path,dst_child_path)
def copy_to_repo(src_path,repo_dir,relative_path,overwrite=True):
    '''copy a file or dir into relative path under repo_dir'''
    dst_path=join_path(repo_dir,relative_path)
    dst_parent = os.path.dirname(dst_path)
    if not os.path.exists(dst_parent):
        os.makedirs(dst_parent)
    if os.path.exists(dst_path):
        if os.path.isdir(src_path):
            dst_path=os.path.join(dst_path,os.path.basename(src_path))
            return shutil.copytree(src_path,dst_path)
        else:
            assert os.path.isfile(src_path)
            if os.path.isdir(dst_path):
                shutil.copy(src_path,os.path.join(dst_path,os.path.basename(src_path)))
            else:
                if overwrite:
                    os.remove(dst_path)
                    shutil.copy(src_path,dst_path)
                else:
                    # this will raise an error, by design.
                    shutil.copy(src_path,dst_path)
                    raise FileExistsError('%s already exists but overwrite is not True'%(dst_path))
    else:
        if os.path.isfile(src_path):
            shutil.copy(src_path,dst_path)
        else:
            assert os.path.isdir(src_path)
            shutil.copytree(src_path,dst_path)







from dulwich import porcelain
from dulwich.repo import Repo
from dulwich.index import build_index_from_tree
from dulwich.client import get_transport_and_path
from dulwich.client import HttpGitClient
import logging
import os
import shutil
import glob

def create_branch(repo,branch,force=False):
    porcelain.branch_create(repo,branch,force=force)
def switch_branch(repo,branch,recover=True):
    porcelain.update_head(repo,branch)
    if recover:
        clean_repo(repo)
        recover_from_head(repo)
def get_current_branch(repo):
    branch=porcelain.active_branch(repo)
    return branch
def is_git_repo(path):
    git_dir=os.path.join(path,'.git')
    return os.path.exists(path) and os.path.exists(git_dir) and os.path.isdir(git_dir)
def commit_current_branch(repo,message='empty message'):
    assert isinstance(repo,Repo)
    repo.do_commit(message=message.encode('utf-8'))
def list_remote_branch(remote_location):
    client, path = get_transport_and_path(remote_location)
    remote_refs,_,_ = client._discover_references(b"git-upload-pack", client._get_url(path))
    branches=[]
    for key in remote_refs:
        key=key.decode()
        if key.startswith('refs/heads/'):
            key=key[11:]
            branches.append(key)
    return branches
def clean_repo(repo):
    path=repo.path
    children=os.listdir(path)
    for child in children:
        child=os.path.join(path,child)
        if os.path.basename(child)=='.git':
            continue
        if os.path.isdir(child):
            shutil.rmtree(child)
        else:
            os.remove(child)
def stage_all_changes(repo):
    status = porcelain.status(repo)
    paths = status.untracked + status.unstaged
    paths = [p.decode() if isinstance(p, bytes) else p for p in paths]
    paths = [repo.path + '/' + p for p in paths]
    if not paths:
        logging.info('Working tree is clean. Nothing to add.')
        return
    paths.append(repo.path)
    porcelain.add(repo, paths)

def recover_from_head(repo):
    '''git checkout -- .'''
    assert isinstance(repo,Repo)
    build_index_from_tree(repo.path,repo.index_path(),repo.object_store,repo[b'HEAD'].tree)
def clone_remote_repo(remote_location,target):
    porcelain.clone(remote_location,target)

def create_head(repo):
    git_add(repo,repo.path)
    commit_current_branch(repo)
def list_branch(repo):
    return [str(x,encoding='utf-8') for x in porcelain.branch_list(repo)]


def delete_branch(repo,branch):
    porcelain.branch_delete(repo,branch)
def merge_branch():
    pass
def fetch_remote_branch():
    pass
def pull_remote_branch(repo,remote_location,branch):
    porcelain.pull(repo,remote_location,branch)
def push_local_branch(repo,branch,remote_location):
   '''local branch must be same with remote branch?'''
   porcelain.push(repo,remote_location,branch)

# def push_local_branch_old(repo,branch,remote_location,remote_branch):
#     client, path = get_transport_and_path(remote_location)
#     remote_branch=('refs/heads/'+remote_branch).encode('utf-8')
#     branch=('refs/heads/'+branch).encode('utf-8')
#     def update_refs(refs):
#         refs[remote_branch] = repo[branch].id
#     client.send_pack(path, update_refs, repo.object_store.generate_pack_data,progress=porcelain.default_bytes_err_stream)

def git_init(path,mkdir=False):
    return Repo.init(path,mkdir)
def git_add(repo,paths):
    if isinstance(paths,str):
        paths=[paths]
    porcelain.add(repo,paths)
def exists_remote_branch(remote_location,branch):
    branches=list_remote_branch(remote_location)
    return branch in branches
def git_stash():
    pass
def git_stash_apply():
    pass
def git_stash_pop():
    pass
def test():
    default_remote_location = 'https://OpenGitspace:Gitspace@123456@gitee.com/OpenGitspace/meta'
    repo_path = './data/OpenGitspace'
    if is_git_repo(repo_path):
        repo = Repo(repo_path)
    else:
        repo = Repo.init(repo_path)

    branches = list_remote_branch(default_remote_location)
    branches = list(filter(lambda x: x[0] == 'n', branches))
    print(branches)
    print('local_branches:',list_branch(repo))
    create_head(repo)
    test_branch='000_test'
    create_branch(repo,test_branch)
    switch_branch(repo,test_branch)
    pull_remote_branch(repo, default_remote_location, 'test')
    push_local_branch(repo,test_branch,default_remote_location,test_branch)
    status = porcelain.status(repo)
    paths = status.untracked + status.unstaged
    print(status)

if __name__ == '__main__':
    test()


from freehub import apis
import os
os.environ['ANSI_COLORS_DISABLED']="1"
import shutil
import fire

class Cli:
    '''
    address: [[[{host}/]{username}/]{repo_name}:]{branch_name}[/{relative_path}]
    '''
    @classmethod
    def download(cls,address,path=None,overwrite=False):
        path=path or './'
        address=apis.get_complete_address(address)
        apis.freehub_download(address,path,overwrite=overwrite)
    @classmethod
    def upload(cls,path,address,overwrite=True):
        address = apis.get_complete_address(address)
        apis.freehub_upload(path,address,overwrite=overwrite)
    @classmethod
    def login(cls):
        apis.freehub_login()
    @classmethod
    def logout(cls):
        apis.freehub_logout()
    @classmethod
    def search(cls,pattern:str):
        pattern=apis.get_complete_address(pattern)
        apis.freehub_search(pattern)

    @classmethod
    def list(cls):
        return cls.search('.*')

def main():
    fire.Fire(Cli)
if __name__ == '__main__':
    main()


import glob
import re
import shutil

from freehub import git_ops
from freehub.utils import *
import logging
import json
import os

# _BRANCH_DICT='branch_dict'
# _FAKE2TRUE: str='fake2true'
# _TRUE2FAKE='true2fake'
# _BRANCH_LIST='remote_branch_list'
USER_HOME=os.path.expanduser('~')
STORE_HOME=USER_HOME+'/.store'
STORE_TMP_DIR='/.store/.tmp'
SCRIPTS_PATH='/.freehub-scripts'
# BRANCH_LIST_DIR=STORE_HOME+'/BranchLists'
# SHADOW_STORE_HOME=STORE_HOME+'/ShadowStores'
# STORE_TMP_DIR=STORE_HOME+'/tmp'
# STORE_CLASS_DIR=STORE_HOME+'/Store.Class'
STORE_CONFIG_FILE=STORE_HOME+'/config.json'
FREEHUB_LOCATION='https://OpenGitspace:Gitspace@123456@gitee.com/OpenGitspace/meta'
DEFAULT_HOST='gitee.com'

DEFAULT_USERNAME='OpenGitspace'
DEFAULT_REPO='gitspace'
DEFAULT_PASSWORD='Gitspace@123456'

def _read_config():
    if not os.path.exists(STORE_CONFIG_FILE):
        return None
    with open(STORE_CONFIG_FILE,'r',encoding='utf-8') as f:
        return json.load(f)
def _write_config(data):
    with open(STORE_CONFIG_FILE,'w',encoding='utf-8') as f:
        json.dump(data,f)

def get_default_repo_address():
    cfg:dict=_read_config()
    username=DEFAULT_USERNAME
    repo=DEFAULT_REPO
    host=DEFAULT_HOST
    if cfg:
        username=cfg.get('username',username)
        host=cfg.get('default_host',host)
        repo=cfg.get('default_repo',repo)
    return '%s/%s/%s'%(host,username,repo)

def get_complete_address(address):
    if ':' in address:
        assert address.count(':') == 1
        repo_address, rest = address.split(':')
        parts=repo_address.split('/')
        assert len(parts)>=1
        assert len(parts)<=3
        default_parts=get_default_repo_address().split('/')
        default_parts[-1]=parts[-1]
        if len(parts)>=2:
            default_parts[-2]=parts[-2]
        if len(parts)>=3:
            default_parts[-3]=parts[-3]
        repo_address='/'.join(default_parts)
    else:
        repo_address = get_default_repo_address()
        rest = address
    addr= ':'.join([repo_address,rest])
    print('Complete address:',addr)
    return addr
def parse_address(address):
    '''
       address: [[[{host}/]{username}/]{repo_name}:]{branch_name}[/{relative_path}]
       '''
    def parse_rest(rest:str):
        rest=rest.strip('/')
        if '/' in rest:
            branch_name,relative_path=rest.split('/',maxsplit=1)
        else:
            branch_name=rest
            relative_path='/'
        return branch_name,relative_path

    assert address.count(':') == 1
    repo_address, rest = address.split(':')
    branch_name, relative_path = parse_rest(rest)
    return repo_address,branch_name,relative_path
def get_repo_url(repo_address:str):
    host,username,repo=repo_address.split('/')
    password=DEFAULT_PASSWORD
    cfg=_read_config()
    if cfg:
        password=cfg.get('password',DEFAULT_PASSWORD)
    repo_url= 'https://%s:%s@%s/%s/%s'%(username,password,host,username,repo)
    # print("repo_url:",repo_url)
    return repo_url




def download_branch(remote_location,branch,dst,overwrite=False):
    check_and_make_empty_dir(dst,overwrite)
    repo=git_ops.git_init(dst)
    git_ops.create_head(repo)
    git_ops.pull_remote_branch(repo,remote_location,branch)

def upload_branch(repo,remote_location,branch,overwrite=False):
    if git_ops.exists_remote_branch(remote_location,branch):
        if not overwrite:
            raise Exception('Remote branch %s already exists at %s .'%(branch,remote_location))
        else:
            logging.warning('Will overwrite remote branch %s at %s .'%(branch,remote_location))
    else:
        logging.info('Will upload to a new branch %s at %s .'%(branch,remote_location))
    branches=git_ops.list_branch(repo)
    if not branch in branches:
        raise Exception('Local branch %s does not exist.'%(branch))
    git_ops.push_local_branch(repo, branch, remote_location)


def download_branch_to_cache(remote_location,branch,cache_dir=None,overwrite=True):
    cache_dir=cache_dir or os.path.join(STORE_HOME,branch)
    check_and_make_empty_dir(cache_dir,overwrite=True)
    download_branch(remote_location,branch,dst=cache_dir,overwrite=overwrite)
    return cache_dir
def download_branch_to_dir(remote_location,branch,dst,cache_dir=None,overwrite=True):
    cache_dir=cache_dir or os.path.join(STORE_HOME,branch)
    download_branch_to_cache(remote_location,branch,cache_dir,overwrite=overwrite)
    copy_repo_files_to_dir(cache_dir,dst)

def upload_to_remote(remote_location,branch_name,relative_path,src_path,cache_dir=None,overwrite=True):
    ''''''
    cache_dir = cache_dir or os.path.join(STORE_HOME, branch_name)
    check_and_make_empty_dir(cache_dir,overwrite=True)
    if git_ops.exists_remote_branch(remote_location, branch_name):
        download_branch_to_cache(remote_location, branch_name, cache_dir)
        repo = git_ops.Repo(cache_dir)
        git_ops.pull_remote_branch(repo, remote_location, branch_name)
    else:
        repo = git_ops.git_init(cache_dir)
    copy_to_repo(src_path,cache_dir,relative_path,overwrite=True)

    #############################

    git_ops.create_head(repo)
    git_ops.create_branch(repo,branch_name,force=True)
    git_ops.switch_branch(repo,branch_name,recover=False)
    git_ops.stage_all_changes(repo)
    git_ops.commit_current_branch(repo)
    upload_branch(repo,remote_location,branch_name,overwrite=overwrite)
def freehub_search(pattern:str):
    '''
    :param reg_str: [[{host}/]{username}/]{repo_name}[:{branch_name}]
    :return:
    '''
    # convert repo_address into repo_url
    repo_address,branch_name_pattern,relative_path=parse_address(pattern)
    repo_url=get_repo_url(repo_address)
    branches=git_ops.list_remote_branch(repo_url)
    for b in branches:
        if branch_name_pattern:
            if re.match(branch_name_pattern,b):
                print(b)
        else:
            print(b)
def freehub_download(address,dst_path,overwrite=False):
    '''
    {cache_dir}/{relative_path}
    relative_path: can be a path pattern, e.g. *.txt
    '''

    repo_address, branch_name,relative_path=parse_address(address)
    repo_url=get_repo_url(repo_address)
    cache_dir=download_branch_to_cache(repo_url,branch_name,overwrite=overwrite)
    copy_repo_files_to(cache_dir,relative_path,dst_path,branch_name)



def freehub_upload(src_path,address,overwrite=False):
    repo_address, branch_name,relative_path=parse_address(address)
    repo_url = get_repo_url(repo_address)
    upload_to_remote(repo_url,branch_name,relative_path,src_path,overwrite=overwrite)

def freehub_run(address):
    ''''''
    raise NotImplementedError

def freehub_login():
    user_name=input('Username:')
    password=input('Password:')
    host=input('Default Host(default host is gitee.com):') or 'gitee.com'
    repo=input('Default Repository:')
    _write_config(dict(
        username=user_name,
        password=password,
        default_repo=repo,
        default_host=host,
    ))
    print('Login successfully.')
def freehub_logout():
    os.remove(STORE_CONFIG_FILE)
    print('Logout successfully.')


def test():
    default_remote_location = FREEHUB_LOCATION
    repo_path = '../data/OpenGitspace'
    import wk
    wk.remake(repo_path)

    # repo_path = './data/'
    # download_branch(default_remote_location,'test',repo_path)
    # upload_branch(default_remote_location,'test2',repo_path,overwrite=True)
    # download_to_dir(default_remote_location,'test',repo_path)
    # upload_to_remote(default_remote_location,'0',repo_path,overwrite=True,cache_dir='data/cache')
    # upload_to_remote(default_remote_location,'0',repo_path,overwrite=True,cache_dir='data/cache')

    wk.Folder(repo_path).open('readme.txt','w').write('yes')
    #freehub
    freehub_upload(repo_path,'0',overwrite=True)
    shutil.rmtree(repo_path)
    freehub_download('0',os.path.dirname(repo_path))
if __name__ == '__main__':
    test()











class NoValue:
    pass
class Path(str):
    __no_value__ = '<__no_value__>'

    def __init__(self, *args, **kwargs):
        super().__init__()

    def __getattr__(self, item):
        return self / item

    def __setattr__(self, key, value):
        self.__dict__[key] = value

    def __truediv__(self, other):
        return Path(self + '/' + other)

    def __call__(self, s=NoValue):
        if s is NoValue: return None
        return self / s

    def relative_path(self, path):
        assert path.startswith(self)
        if self == '':
            pass
        else:
            if path.startswith(self):
                path = path[len(self):]
            path = path.lstrip('/')
        return self.__class__(path)

class StrictPath:
    def __init__(self, s):
        self.__value__ = Path(self.__strict__(s))

    def __strict__(self, s):
        prefix = '/' if s.startswith('/') or s.startswith("\\") else ''

        def remove_all(lis, item):
            if item in lis:
                lis2 = lis.copy()
                for i in lis:
                    if i == item:
                        lis2.remove(i)
                return lis2

            else:
                return lis

        lis = s.split('/')
        lis2 = []
        for i in lis:
            lis2 += i.split('\\')
        lis = lis2
        lis = remove_all(lis, '/')
        lis = remove_all(lis, "\\")
        lis = remove_all(lis, '')
        return prefix + '/'.join(lis)

    def __getattr__(self, item):
        return self / item

    def __truediv__(self, other):
        return StrictPath(self.__value__ / other).__value__

    def __call__(self, s=''):
        if s == '': return self.__value__
        return StrictPath(self.__value__ / s)

    def __repr__(self):
        return "<StrictPath:'%s'>" % (self.__value__)

    def __str__(self):
        return self.__value__


def join_path(*args):
    path = StrictPath('/'.join(args))()
    path = standard_path(path)
    return path


def standard_path(p, check=False):
    assert len(p)
    # if (len(p)<=4 and p[1:]=='://') or (len(p)==3 and p[1:]==':/'):return p[:3]
    # print(p)
    p = str(StrictPath(p))
    # print(p)
    # if not '/' in p:return p
    p = p.split('/')
    assert len(p)
    res = []
    p.reverse()
    while True:
        if not len(p):
            if len(res) > 1 and '.' in res:
                for char in res:
                    if char == '.':
                        res.remove(char)
            path = '/'.join(res)
            if len(path) == 2 and path[1] == ':':
                path = path + '/'
            return path
        i = p.pop()
        if i == '':
            res.append('')
            continue
        elif i == '.':
            if not len(res): res.append('.')
        elif i == '..':
            if check: return False
            if not len(res) or res[-1] == '.':
                raise Exception('Error,path has reached the top when another ".." shew up.')
            else:
                res.pop()
                if not len(res): res.append('.')
        else:
            res.append(i)

def get_relative_path(root, child):
    path = child
    path = standard_path(path)
    root = standard_path(root)
    assert path.startswith(root)
    return path[len(root):]